           (*-------------------------------------------------*
            |                 (a part of) ep2                 |
            |                  September 2004                 |
            |                   December 2004 (modified)      |
            |  Markus Roggenbach (Univ of Wales Swansea, UK)  |
            |  Yoshinao Isobe    (AIST, Japan)                |
            *-------------------------------------------------*)

theory ep2_nucleus = CSP_Prover:

(*** To automatically unfold syntactic sugar ***)
declare csp_ss_def[simp]

(*********************************************************
              data type passed on channels
 *********************************************************)

typedecl init_d typedecl request_d typedecl response_d typedecl exit_d

datatype Data = Init    init_d    | Exit     exit_d
              | Request request_d | Response response_d

(*********************************************************
                     event (channel)
 *********************************************************)

datatype Event = c Data

(*********************************************************
         abstract component description level
 *********************************************************)

datatype ACName = Acquirer | AcConfigManagement 
                | Terminal | TerminalConfigManagement

consts 
  ACDef :: "(ACName, Event) procDF"

primrec
  "ACDef (Acquirer) 
         = c ? x:(range Init) -> <AcConfigManagement>"

  "ACDef (AcConfigManagement)
         = c !! exit:(range Exit) -> SKIP
           |~| 
           c !! request:(range Request)
             -> c ? response:(range Response)
             -> <AcConfigManagement>"

  "ACDef (Terminal)
         = c !! init:(range Init) -> <TerminalConfigManagement>"

  "ACDef (TerminalConfigManagement)
         = c ? x -> IF (x:range Request) 
                    THEN c !! response:(range Response)
                          -> <TerminalConfigManagement>
                    ELSE IF (x:range Exit)
                         THEN SKIP
                         ELSE STOP"

constdefs
  AC :: "(ACName, Event) procRC"
        "AC ==  LET ACDef
                IN (<Acquirer> |[range c]| <Terminal>)"

(*---------------------------------------------------------------*
 |                          NOTE                                 |
 |                                                               |
 | c ! v -> P       : sends a value v to c, then behaves like P. |
 |                                                               |
 | c ? x:X -> P(x)  : receives a value v from c if c in X,       |
 |                    then behaves like P(v).                    |
 |                                                               |
 | c !! x:X -> P(x) : sends a value v selected from X to c,      |
 |                    then behaves like P(v).                    |
 |                                                               |
 *---------------------------------------------------------------*)

(*********************************************************
                  architectual level
 *********************************************************)

datatype AbsName = Abstract | Loop

consts 
  AbsDef :: "(AbsName, Event) procDF"

primrec
  "AbsDef (Abstract)
          = c !! init:(range Init) -> <Loop>"

  "AbsDef (Loop)
          = c !! exit:(range Exit) -> SKIP
            |~| 
            c !! request:(range Request)
              -> c !! response:(range Response)
              -> <Loop>"

consts
  Abs :: "(AbsName, Event) procRC"

defs
  Abs_def:
   "Abs == 
     LET AbsDef IN <Abstract>"

(*********************************************************
        relating function between ACName and AbsName
 *********************************************************)

consts 
  Abs_to_AC :: "AbsName => (ACName, Event) proc"

primrec
  "Abs_to_AC (Abstract)
          = (<Acquirer> |[range c]| <Terminal>)"

  "Abs_to_AC (Loop)
          = (<AcConfigManagement> |[range c]| <TerminalConfigManagement>)"

(*********************************************************
        nohide and guarded lemmas (routine work)
 *********************************************************)

lemma guard_nohide[simp]: 
  "!! C. guard (ACDef  C) & nohide (ACDef  C)"
  "!! C. guard (AbsDef C) & nohide (AbsDef C)"
by (induct_tac C, simp_all, induct_tac C, simp_all)

(*********************************************************
           a theorem for verifying Abs =F AC
 *********************************************************)

theorem ep2: "Abs =F AC"
apply (unfold Abs_def AC_def)
apply (rule csp_fp_induct_cms[of _ _ _ "Abs_to_AC"], simp_all)
apply (induct_tac C)
by (auto simp add: image_iff | tactic {* csp_hnf_tac 1 *} | rule csp_decompo)+

(*--------------------------------------------------------*
 |                                                        |
 |        another proof of Abs =F AC (more manual)        |
 |                                                        |
 |           The proof is divided to two lemmas.          |
 |                                                        |
 *--------------------------------------------------------*)

declare csp_ss_def [simp del]

(*** initializing ***)

lemma ep2_step1:
   "LET ACDef IN Rewrite (AbsDef Abstract) By Abs_to_AC =F 
    LET ACDef IN Abs_to_AC Abstract"
apply (tactic {* csp_unwind_tac 1 *})
apply (unfold csp_ss_def, tactic {* csp_dist_tac 1 *})
apply (rule csp_decompo, simp)

apply (tactic {* csp_step_tac 1 *})
apply (auto)
apply (tactic {* csp_simp_tac 1 *})
done

(*** main loop ***)

lemma ep2_step2:
   "LET ACDef IN Rewrite (AbsDef Loop) By Abs_to_AC =F 
    LET ACDef IN Abs_to_AC Loop"
apply (tactic {* csp_unwind_tac 1 *})
apply (unfold csp_ss_def, tactic {* csp_dist_tac 1 *})
apply (rule csp_decompo)

(* case of eixt *)
apply (rule csp_decompo, simp)
apply (tactic {* csp_step_tac 1 *})
apply (auto simp add: image_iff)
apply (tactic {* csp_step_tac 1 *})

(* case of loop *)

 (* request phase *)
apply (rule csp_decompo, simp)
apply (tactic {* csp_step_tac 1 *})
apply (auto simp add: image_iff)
apply (tactic {* csp_simp_tac 1 *})

 (* response phase *)
apply (tactic {* csp_dist_tac 1 *})
apply (rule csp_decompo, simp)
apply (tactic {* csp_step_tac 1 *})
apply (auto simp add: image_iff)
apply (tactic {* csp_simp_tac 1 *})
done

(**********************
       Abs =F AC ?
 **********************)

theorem ep2_man: "Abs =F AC"
apply (simp add: Abs_def AC_def)

  (***** by fixed point induction *****)
apply (rule csp_fp_induct_cms[of _ _ _ "Abs_to_AC"])

  (***** check guarded and no hiding operators *****)
apply (simp_all)
apply (induct_tac C)
apply (simp add: ep2_step1 del: AbsDef.simps Abs_to_AC.simps)
apply (simp add: ep2_step2 del: AbsDef.simps Abs_to_AC.simps)
done

end
